/*
Copyright (C) 2013 Bengt Martensson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see http://www.gnu.org/licenses/.
 */

package org.harctoolbox.irscrutinizer;

import com.beust.jcommander.JCommander;
import com.beust.jcommander.Parameter;
import com.beust.jcommander.ParameterException;
import com.hifiremote.exchangeir.RepeatFinder;
import gnu.io.NoSuchPortException;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;
import java.net.URISyntaxException;
import java.net.URL;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Locale;
import javax.xml.parsers.ParserConfigurationException;
import org.harctoolbox.IrpMaster.DecodeIR;
import org.harctoolbox.IrpMaster.ExchangeIR;
import org.harctoolbox.IrpMaster.IncompatibleArgumentException;
import org.harctoolbox.IrpMaster.IrSignal;
import org.harctoolbox.IrpMaster.IrpMasterException;
import org.harctoolbox.IrpMaster.IrpUtils;
import org.harctoolbox.IrpMaster.ModulatedIrSequence;
import org.harctoolbox.guicomponents.GuiUtils;
import org.harctoolbox.harchardware.HarcHardwareException;
import org.harctoolbox.harchardware.ir.IrWidget;
import org.harctoolbox.harchardware.comm.LocalSerialPort;
import org.xml.sax.SAXException;

/**
 * This class decodes command line parameters and fires up the GUI.
 */
public class IrScrutinizer {

    /** Number indicating invalid value. */
    public final static long invalid = -1;

    private IrScrutinizer() {
    }

    private static void usage(int exitcode) {
        StringBuilder str = new StringBuilder();
        argumentParser.usage(str);

        (exitcode == IrpUtils.exitSuccess ? System.out : System.err).println(str);
        doExit(exitcode); // placifying FindBugs...
    }

    private static void doExit(int exitcode) {
        System.exit(exitcode);
    }

    private final static class CommandLineArgs {
        @Parameter(names = {"-b", "--begintimeout"}, description = "Timeout in milliseconds for start")
        private int beginTimeout = 2000;

        @Parameter(names = {"-c", "--commandline"}, description = "Run in command line mode")
        private boolean commandLine = false;

        @Parameter(names = {"-g", "--gui"}, description = "Run in GUI mode")
        private boolean gui = true;

        @Parameter(names = {"-d", "--device"}, description = "Device name")
        private String deviceName = System.getProperty("os.name").startsWith("Windows") ? "COM1:" : "/dev/ttyUSB0";

        @Parameter(names = {"-D", "--debug"}, description = "Debug code")
        private int debug = 0;

        @Parameter(names = {"-e", "--endtimeout"}, description = "Timeout in milliseconds for ending")
        private int endTimeout = 500;

        @Parameter(names = {"-h", "--help", "-?"}, description = "Display help message")
        private boolean helpRequested = false;

        @Parameter(names = {"-H", "--home", "--applicationhome", "--apphome"}, description = "Set application home (where files are located)")
        private String applicationHome = null;

        @Parameter(names = {"-l", "--length"}, description = "Max length in milliseconds of recorded sequence")
        private int length = 10000;

        @Parameter(names = {"-L", "--listports"}, description = "List serial ports")
        private boolean listPorts = false;

        //@Parameter(names = {"-T", "--timeout"}, description = "Timeout in milliseconds")
        //private int timeout = 2000;

        @Parameter(names = {"-V", "--version"}, description = "Display version information")
        private boolean versionRequested;

        @Parameter(names = {"-v", "--verbose"}, description = "Have some commands executed verbosely")
        private boolean verbose;

        @Parameter(names = {"-#", "--count"}, description = "Number of times to send sequence")
        private int count = 1;

        //@Parameter(names = {"-c", "--config"}, description = "Path to IrpProtocols.ini")
        //private String irprotocolsIniFilename = "data/IrpProtocols.ini";

        //@Parameter(names = {"--getversion"}, description = "Call the getVersion() function")
        //private boolean getversion = false;

        //@Parameter(names = {"-i", "--ip"}, description = "IP address or name")
        //private String ip = null;

        //@Parameter(names = {"-I", "--irtrans"}, description = "Use IrTrans")
        //private boolean irtrans = false;

        //@Parameter(names = {"-l", "--lirc"}, description = "Use Lirc Client")
        //private boolean lirc = false;

        @Parameter(names = {"-p", "--properties"}, description = "Pathname of properties file")
        private String propertiesFilename = null;

        //@Parameter(names = {"-s", "--sendircommand"}, arity = 2, description = "Send a preprogrammed IR command by name of remote and command")
        //private List<String> remotecommand = null;

        //@Parameter(names = {"-t", "--transmitter"}, description = "Transmitter, semantic device dependent")
        //private String transmitter = null;

        //@Parameter(names = {"-T", "--timeout"}, description = "Timeout in milliseconds")
        //private int timeout = 2000;

        //@Parameter(names = {"-y", "--irtoy"}, description = "Use IrToy")
        //private boolean irtoy = false;

        //@Parameter(description = "[parameters]")
        //private ArrayList<String> parameters = new ArrayList<String>();
    }

    private static JCommander argumentParser;
    private static CommandLineArgs commandLineArgs = new CommandLineArgs();

    /**
     * @param args the command line arguments.
     */
    public static void main(String[] args) {
        int userLevel = 0; // presently not used
        argumentParser = new JCommander(commandLineArgs);
        argumentParser.setProgramName(Version.appName);

        try {
            argumentParser.parse(args);
        } catch (ParameterException ex) {
            System.err.println(ex.getMessage());
            usage(IrpUtils.exitUsageError);
        }

        if (commandLineArgs.helpRequested)
            usage(IrpUtils.exitSuccess);

        if (commandLineArgs.versionRequested) {
            System.out.println(Version.versionString);
            System.out.println("JVM: " + System.getProperty("java.vendor") + " " + System.getProperty("java.version") + " " + System.getProperty("os.name") + "-" + System.getProperty("os.arch"));
            System.out.println();
            System.out.println(Version.licenseString);
            System.exit(IrpUtils.exitSuccess);
        }

        String applicationHome =  commandLineArgs.applicationHome != null ? commandLineArgs.applicationHome : System.getenv("IRSCRUTINIZERHOME");
        if (applicationHome == null) {
            URL url = IrScrutinizer.class.getProtectionDomain().getCodeSource().getLocation();
            File dir = null;
            try {
                dir = new File(url.toURI()).getParentFile();
            } catch (URISyntaxException ex) {
                System.err.println(ex.getMessage());
                System.exit(IrpUtils.exitFatalProgramFailure);
            }
            if (dir == null) {
                System.err.println("Error getting applicationHome");
                System.exit(IrpUtils.exitFatalProgramFailure);
            } else
                applicationHome = (dir.getName().equals("build") || dir.getName().equals("dist"))
                    ? dir.getParent() : dir.getPath();
        }
        if (applicationHome != null && !applicationHome.endsWith(File.separator))
            applicationHome += File.separator;

        if (commandLineArgs.debug > 0)
            System.err.println("applicationHome = " + applicationHome);


        String libPath = System.getProperty("java.library.path");
        String libraries = applicationHome
                + (System.getProperty("os.name").startsWith("Windows") ? "Windows" : System.getProperty("os.name"))
                + '-' + System.getProperty("os.arch").toLowerCase(Locale.US)
                + File.pathSeparatorChar + libPath;
        // see http://blog.cedarsoft.com/2010/11/setting-java-library-path-programmatically/
        System.setProperty("java.library.path", libraries);
        if (commandLineArgs.debug > 0)
            System.err.println("java.library.path = " + libraries);
        try {
            Field fieldSysPath = ClassLoader.class.getDeclaredField("sys_paths");
            fieldSysPath.setAccessible(true);
            fieldSysPath.set(null, null);
        } catch (NoSuchFieldException ex) {
            System.err.println(ex.getMessage());
            System.exit(IrpUtils.exitFatalProgramFailure);
        } catch (SecurityException ex) {
            System.err.println(ex.getMessage());
            System.exit(IrpUtils.exitFatalProgramFailure);
        } catch (IllegalArgumentException ex) {
            System.err.println(ex.getMessage());
            System.exit(IrpUtils.exitFatalProgramFailure);
        } catch (IllegalAccessException ex) {
            System.err.println(ex.getMessage());
            System.exit(IrpUtils.exitFatalProgramFailure);
        }

        try {
            if (commandLineArgs.listPorts) {
                ArrayList<String> names = LocalSerialPort.getSerialPortNames(false);
                for (String port : names) {
                    System.out.println(port);
                }
            }

            // commandLine overrides gui
            if (commandLineArgs.commandLine)
                commandLineExecute(commandLineArgs.deviceName, commandLineArgs.beginTimeout,
                        commandLineArgs.length, commandLineArgs.endTimeout, commandLineArgs.count, commandLineArgs.verbose, commandLineArgs.debug);
            else if (commandLineArgs.gui)
                guiExecute(applicationHome, commandLineArgs.propertiesFilename, commandLineArgs.verbose, commandLineArgs.debug, userLevel);
        } catch (NoSuchPortException ex) {
            System.err.println(ex.getMessage());
            System.exit(IrpUtils.exitIoError);
        } catch (HarcHardwareException ex) {
            System.err.println(ex.getMessage());
            System.exit(IrpUtils.exitIoError);
        } catch (IOException ex) {
            System.err.println(ex.getMessage());
            System.exit(IrpUtils.exitIoError);
        } catch (IrpMasterException ex) {
            System.err.println(ex.getMessage());
            System.exit(IrpUtils.exitIoError);
        }
    }

    private static void commandLineExecute(String deviceName, int beginTimeout,
            int length, int endTimeout, int count, boolean verbose, int debug) throws NoSuchPortException, HarcHardwareException, IOException, IrpMasterException {
        IrWidget irWidget = null;
        irWidget = new IrWidget(deviceName, beginTimeout, length, endTimeout, verbose);

        for (int i = 0; i < count; i++) {
            ModulatedIrSequence seq = null;
                seq = irWidget.capture();
                //System.err.println(success);
                //seq = success ? irWidget.getIrSequence() : null;
                System.err.println(seq);


            if (seq == null) {
                System.err.println("No input");
                //System.exit(IrpUtils.exitIoError);
                continue;
            }
            if (verbose)
                System.out.println(seq);
            RepeatFinder repeatFinder = ExchangeIR.newRepeatFinder(seq);
            if (commandLineArgs.verbose)
                System.out.println(repeatFinder);
            IrSignal signal = ExchangeIR.interpretIrSequence(seq, true);
            if (commandLineArgs.verbose)
                System.out.println(signal);
            DecodeIR.invoke(signal);
        }
        irWidget.close();
    }

    private static void guiExecute(final String applicationHome, final String propsfilename, final boolean verbose, final int debug, final int userlevel) {
        java.awt.EventQueue.invokeLater(new Runnable() {

            @Override
            public void run() {
                try {
                    new GuiMain(applicationHome, propsfilename, verbose, debug, userlevel).setVisible(true);
                } catch (ParseException ex) {
                    GuiUtils.fatal(ex, IrpUtils.exitConfigReadError);
                } catch (ParserConfigurationException ex) {
                    GuiUtils.fatal(ex, IrpUtils.exitInternalFailure);
                } catch (SAXException ex) {
                    GuiUtils.fatal(ex, IrpUtils.exitXmlError);
                } catch (IOException ex) {
                    GuiUtils.fatal(ex, IrpUtils.exitConfigReadError);
                } catch (IncompatibleArgumentException ex) {
                    GuiUtils.fatal(ex, IrpUtils.exitConfigReadError);
                } catch (URISyntaxException ex) {
                    GuiUtils.fatal(ex, IrpUtils.exitConfigReadError);
                }
            }
        });
    }
}
